//  ===================================================================
//
//  Library:   commonlib
//  Module:    3dvector.h
//  Language:  C++
//  Date:      June 2004
//  Version:   1.0.0
//
//  Copyright (c) 2004
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
//  ==================================================================

#ifndef __3dvector_h_
#define __3dvector_h_

#include <sstream>

#include "mathmacros.h"
#include "exceptionobj.h"


namespace common { 


// -------------------------------------------------------------------
// class t3DVector
//
// This class represents vector in the three-dimensional space. 
//
// -------------------------------------------------------------------

class t3DVector {
public:
  // constructors
  t3DVector() {}
  t3DVector(const t3DVector&);
  t3DVector(double, double, double);

  // public methods
  void setCoordinates(const t3DVector&);
  void setCoordinates(double, double, double);

  bool       operator ==(const t3DVector&) const;
  bool       operator !=(const t3DVector&) const;
  t3DVector& operator  =(const t3DVector&);
  t3DVector& operator +=(const t3DVector&);
  t3DVector& operator -=(const t3DVector&);
  t3DVector& operator *=(double);
  t3DVector  operator  -() const;
  t3DVector  operator  +(const t3DVector&) const;
  t3DVector  operator  -(const t3DVector&) const;
  t3DVector  operator  *(double) const;

  friend t3DVector operator *(double, const t3DVector&);

  double  operator  *(const t3DVector&) const;
  double  operator [](int) const throw (tExceptionObject);
  double& operator [](int) throw (tExceptionObject);

  double  length() const;
  bool	  isNull() const;

  t3DVector& negate();
  t3DVector& normalize();

  double    inner(const t3DVector&) const;
  t3DVector	cross(const t3DVector&) const;

  // public attributes (to speed up the access)
  double m_x;
  double m_y;
  double m_z;
};


// -------------------------------------------------------------------
// inline implementation
// -------------------------------------------------------------------
inline
t3DVector::t3DVector(const t3DVector& vect) :
  m_x(vect.m_x),
  m_y(vect.m_y),
  m_z(vect.m_z)
{}

inline
t3DVector::t3DVector(double ax, double ay, double az) :
  m_x(ax),
  m_y(ay),
  m_z(az)
{}

inline void
t3DVector::setCoordinates(const t3DVector& vect)
{
  m_x = vect.m_x;
  m_y = vect.m_y;
  m_z = vect.m_z;
}

inline void
t3DVector::setCoordinates(double ax, double ay, double az)
{
  m_x = ax;
  m_y = ay;
  m_z = az;
}

inline bool
t3DVector::operator ==(const t3DVector& vect) const
{
  return isEqual(m_x, vect.m_x) && isEqual(m_y, vect.m_y) &&
  isEqual(m_z, vect.m_z);
}

inline bool
t3DVector::operator !=(const t3DVector& vect) const
{
  return !operator ==(vect);
}

inline t3DVector&
t3DVector::operator =(const t3DVector& vect)
{
  m_x = vect.m_x;
  m_y = vect.m_y;
  m_z = vect.m_z;

  return *this;
}

inline t3DVector&
t3DVector::operator +=(const t3DVector& vect)
{
  m_x += vect.m_x;
  m_y += vect.m_y;
  m_z += vect.m_z;

  return *this;
}

inline t3DVector&
t3DVector::operator -=(const t3DVector& vect)
{
  m_x -= vect.m_x;
  m_y -= vect.m_y;
  m_z -= vect.m_z;

  return *this;
}

inline t3DVector&
t3DVector::operator *=(double alpha)
{
  m_x *= alpha;
  m_y *= alpha;
  m_z *= alpha;

  return *this;
}

inline t3DVector
t3DVector::operator -() const
{
  return t3DVector(-m_x, -m_y, -m_z);
}

inline t3DVector
t3DVector::operator +(const t3DVector& vect) const
{
  return t3DVector(
    m_x + vect.m_x, 
    m_y + vect.m_y, 
    m_z + vect.m_z
  );
}

inline t3DVector
t3DVector::operator -(const t3DVector& vect) const
{
  return t3DVector(
    m_x - vect.m_x, 
    m_y - vect.m_y, 
    m_z - vect.m_z
  );
}

inline t3DVector
t3DVector::operator *(double alpha) const
{
  return t3DVector(
    m_x * alpha, 
    m_y * alpha, 
    m_z * alpha
  );
}

inline t3DVector
operator *(double alpha, const t3DVector& vect)
{
  return vect * alpha;
}

inline double
t3DVector::operator *(const t3DVector& vect) const
{
  return m_x * vect.m_x + m_y * vect.m_y + m_z * vect.m_z;
}

inline double
t3DVector::operator [](int i) const throw (tExceptionObject) {
  if ((i < 0) || (i > 2)) {
    std::stringstream ss (std::stringstream::in | 
    std::stringstream::out);
    ss << "Array index is out of range";
    throw tExceptionObject(__FILE__,__LINE__,ss.str());  
  }

  if (i == 0) {
    return m_x;
  }
  else if (i == 1) {
    return m_y;
  }

  return m_z;
}

inline double&
t3DVector::operator [](int i) throw (tExceptionObject) {
  if ((i < 0) || (i > 2)) {
    std::stringstream ss (std::stringstream::in | 
    std::stringstream::out);
    ss << "Array index is out of range";
    throw tExceptionObject(__FILE__,__LINE__,ss.str());
  }

  if (i == 0) {
    return m_x;
  }
  else if (i == 1) {
    return m_y;
  }

  return m_z;
}

inline double
t3DVector::length() const {
  return sqrt(sqr(m_x) + sqr(m_y) + sqr(m_z));
}

inline bool
t3DVector::isNull() const {
  return isZero(m_x) && isZero(m_y) && isZero(m_z);
}

inline t3DVector&
t3DVector::negate() {
  m_x = -m_x;
  m_y = -m_y;
  m_z = -m_z;

  return *this;
}

inline double
t3DVector::inner(const t3DVector& vect) const {
  return operator *(vect);
}


}         // end of namespace common


#endif     // __3dvector_h_
