//  ===================================================================
//
//  Program:  commonlib
//  Module:   svddcmp.cpp
//  Language: C++
//  Date:     February 2005
//  Version:  1.0.0
//
//  Copyright (c) 2005
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
//  ==================================================================

#include "svddcmp.h"

#include "macros.h"
#include "mathmacros.h"

#include <sstream>
#include <stdlib.h>
#include <cmath>


namespace common {


  // -----------------------------------------------------------------
  // Method: tSVDdcmp()
  // 
  // Default constructor
  //
  // -----------------------------------------------------------------
  tSVDdcmp::tSVDdcmp(
                     double** a,
                     int m,
                     int n
                    )
  {
    // initialize dimensions of matrix "matA".
    m_dimm = m;
    m_dimn = n;

    // allocate memory for the  matrices "matA" and "matV", and vector
    // "vecW".
    allocate();

    // copy data from matrix "a"  to matrix "matA", and fill in "vecW"
    // with zeros.
    int i, j;
    for (i=0; i<m_dimm; i++) {
      for (j=0; j<m_dimn; j++) {
        m_matA[i][j] = a[i][j];
      }
    }

    // fill in the matrix "matV" and vector "vecW" with zeros.
    for (i=0; i<m_dimn; i++) {
      m_vecW[i] = 0;
      for (j=0; j<m_dimn; j++) {
        m_matV[i][j] = 0;
      }
    }

    // compute the SVD decomposition of matrix "matA".
    decomp();

    // find the largest singular value of "matA".
    double wmax = 0;
    for (i=0; i<m_dimn; i++) {
      if (m_vecW[i] > wmax) wmax = m_vecW[i];
    }

    // compute  the  threshold  for  singular  values  allowed  to  be
    // nonzero.
    double wmin = wmax * 1.0e-12;
    for (i=0; i<m_dimn; i++) {
      if (m_vecW[i] < wmin) m_vecW[i] = 0;
    }
  }


  // -----------------------------------------------------------------
  // Method: tSVDdcmp()
  // 
  // Copy constructor
  //
  // -----------------------------------------------------------------
  tSVDdcmp::tSVDdcmp(const tSVDdcmp& s)
  {
    // initialize dimensions of matrix "matA"
    m_dimm = s.m_dimm;
    m_dimn = s.m_dimn;

    // allocate memory for the  matrices "matA" and "matV", and vector
    // "vecW".
    allocate();

    int i, j;
    for (i=0; i<m_dimm; i++) {
      for (j=0; j<m_dimn; j++) {
        m_matA[i][j] = s.m_matA[i][j];
      }
    }

    for (i=0; i<m_dimn; i++) {
      m_vecW[i] = s.m_vecW[i];
      for (j=0; j<m_dimn; j++) {
        m_matV[i][j] = s.m_matV[i][j];
      }
    }
  }


  // -----------------------------------------------------------------
  // Method: ~tSVDdcmp()
  // 
  // Destructor
  //
  // -----------------------------------------------------------------
  tSVDdcmp::~tSVDdcmp() 
  {
    // release memory  allocated for  the matrices "matA"  and "matV",
    // and vector "vecW".
    deallocate();
  }


  // -----------------------------------------------------------------
  // Method: matA()
  // 
  // Returns the (i,j) element of the matrix "matA"
  //
  // -----------------------------------------------------------------
  double
  tSVDdcmp::matA(
                 int i,
                 int j
                )
    const throw (tExceptionObject)
  {
    if ((i < 0) || (i >= m_dimm) || (j < 0) || (j >= m_dimn)) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "Bad element index";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    return m_matA[i][j];
  }


  // -----------------------------------------------------------------
  // Method: matV()
  // 
  // Returns the (i,j) element of the matrix "matV"
  //
  // -----------------------------------------------------------------
  double
  tSVDdcmp::matV(
                 int i,
                 int j
                )
    const throw (tExceptionObject)
  {
    if ((i < 0) || (i >= m_dimn) || (j < 0) || (j >= m_dimn)) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "Bad element index";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    return m_matV[i][j];
  }


  // -----------------------------------------------------------------
  // Method: vecW()
  // 
  // Returns the i-th element of the vector "vecW".
  //
  // -----------------------------------------------------------------
  double
  tSVDdcmp::vecW(int i) const throw (tExceptionObject)
  {
    if ((i < 0) || (i >= m_dimn)) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "Bad element index";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    return m_vecW[i];
  }


  // -----------------------------------------------------------------
  // Method: lsqSolve()
  // 
  // Use SVD to solve a linear least squares problem ||Ax - B||.
  //
  // -----------------------------------------------------------------
  void
  tSVDdcmp::lsqSolve(
                     double* b,
	             double* x
                    )
    const
  {
    // compute Ax - B = 0 by backsubstitution.
    bksb(b,x);
  }


  // -----------------------------------------------------------------
  // Method: allocate()
  // 
  // Allocate memory for "matA", "matV", and "vecW".
  //
  // -----------------------------------------------------------------
  void
  tSVDdcmp::allocate()
  {
    m_matA = (double **) new double*[m_dimm];
    m_matV = (double **) new double*[m_dimn];
    m_vecW = (double  *) new double[m_dimn];

    int i;
    for (i=0; i<m_dimm; ++i) {
      m_matA[i] = (double*) new double[m_dimn];
    }

    for (i=0; i<m_dimn; ++i) {
      m_matV[i] = (double*) new double[m_dimn];
    }
  }


  // -----------------------------------------------------------------
  // Method: deallocate()
  // 
  // Release memory used by this quadric matrix.
  //
  // -----------------------------------------------------------------
  void
  tSVDdcmp::deallocate()
  {
    if (m_matA != 0) {
      int i;
      for (i=0; i<m_dimm; ++i) {
        if (m_matA[i] != 0) delete[] m_matA[i];
      }

      delete[] m_matA;
    }

    if (m_matV != 0) {
      int i;
      for (i=0; i<m_dimn; ++i) {
        if (m_matV[i] != 0) delete[] m_matV[i];
      }

      delete[] m_matV;
    }

    if (m_vecW) delete[] m_vecW;

    m_matA = 0;
    m_matV = 0;
    m_vecW = 0;
  }


  // -----------------------------------------------------------------
  // Method: decomp()
  // 
  // Computes the SVD  of a M x N  matrix A. This method is  a copy of
  // the procedure  "svdcmp" from the  book Numerical Recipes  in C++:
  // The  Art of Scientific  Computing, by  W.  Press,  S.  Teukolsky,
  // W.  Vetterling,  and  B.  Flannery,  Secoond  Edition,  Cambridge
  // University Press, 2002.
  //
  // -----------------------------------------------------------------
  void
  tSVDdcmp::decomp() const throw (tExceptionObject)
  {
    bool flag;
    int i,its,j,jj,k,l=0,nm=0;
    double anorm,c,f,g,h,s,scale,x,y,z;

    double* rv1 = (double*) new double[m_dimn];
    g=scale=anorm=0.0;
    for (i=0; i<m_dimn; i++) {
      l=i+2;
      rv1[i]=scale*g;
      g=s=scale=0.0;
      if (i < m_dimm) {
        for (k=i; k<m_dimm; k++) {
          scale += fabs(m_matA[k][i]);
        }
	if (scale != 0.0) {
          for (k=i; k<m_dimm; k++) {
	    m_matA[k][i] /= scale;
            s += m_matA[k][i]*m_matA[k][i];
          }

          f=m_matA[i][i];
          g = -sign(sqrt(s),f);
          h=f*g-s;
          m_matA[i][i]=f-g;
	  
          for (j=l-1; j<m_dimn; j++) {
            for (s=0.0, k=i; k<m_dimm; k++) {
              s += m_matA[k][i] * m_matA[k][j];
            }

            f=s/h;

            for (k=i; k<m_dimm; k++) {
              m_matA[k][j] += f * m_matA[k][i];
            }
	  }

          for (k=i; k<m_dimm; k++) {
            m_matA[k][i] *= scale;
          }
        }
      }

      m_vecW[i]=scale * g;
      g =s=scale=0.0;

      if ((i+1 <= m_dimm) && (i+1 != m_dimn)) {
        for (k=l-1; k<m_dimn; k++) {
          scale += fabs(m_matA[i][k]);
        }
        if (scale != 0.0) {
          for (k=l-1; k<m_dimn; k++) {
            m_matA[i][k] /= scale;
            s += m_matA[i][k] * m_matA[i][k];
          }

          f=m_matA[i][l-1];
          g = -sign(sqrt(s),f);
          h=f*g-s;
          m_matA[i][l-1]=f-g;

          for (k=l-1; k<m_dimn; k++) {
            rv1[k] = m_matA[i][k] / h;
          }

	  for (j=l-1; j<m_dimm; j++) {
	    for (s=0.0, k=l-1; k<m_dimn; k++) {
              s += m_matA[j][k] * m_matA[i][k];
            }
	    for (k=l-1; k<m_dimn; k++) {
              m_matA[j][k] += s * rv1[k];
            }
	  }

          for (k=l-1; k<m_dimn; k++) {
            m_matA[i][k] *= scale;
          }
        }
      }
      anorm=MAX(anorm,(fabs(m_vecW[i])+fabs(rv1[i])));
    }

    for (i=m_dimn-1; i>=0; i--) {
      if (i < m_dimn-1) {
        if (g != 0.0) {
          for (j=l; j<m_dimn; j++) {
            m_matV[j][i] = (m_matA[i][j] / m_matA[i][l]) / g;
          }

          for (j=l; j<m_dimn; j++) {
            for (s=0.0, k=l; k<m_dimn; k++) {
              s += m_matA[i][k] * m_matV[k][j];
            }
            for (k=l; k<m_dimn; k++) {
              m_matV[k][j] += s * m_matV[k][i];
            }
          }
        }

        for (j=l; j<m_dimn; j++) {
          m_matV[i][j] = m_matV[j][i] = 0.0;
        }
      }

      m_matV[i][i]=1.0;
      g=rv1[i];
      l=i;
    }

    for (i=MIN(m_dimm,m_dimn)-1; i>=0; i--) {
      l=i+1;
      g=m_vecW[i];
      for (j=l; j<m_dimn; j++) {
        m_matA[i][j]=0.0;
      }

      if (g != 0.0) {
        g=1.0/g;
        for (j=l; j<m_dimn; j++) {
          for (s=0.0, k=l; k<m_dimm; k++) {
            s += m_matA[k][i] * m_matA[k][j];
          }

          f=(s/m_matA[i][i])*g;

          for (k=i; k<m_dimm; k++) {
            m_matA[k][j] += f * m_matA[k][i];
          }
	}

        for (j=i; j<m_dimm; j++) {
          m_matA[j][i] *= g;
        }
      } 
      else {
        for (j=i; j<m_dimm; j++) {
          m_matA[j][i]=0.0;
        }
      }

      ++m_matA[i][i];
    }

    for (k=m_dimn-1; k>=0; k--) {
      for (its=0; its<30; its++) {
        flag=true;
        for (l=k; l>=0; l--) {
          nm=l-1;
          if ((fabs(rv1[l]) + anorm) == anorm) {
            flag=false;
            break;
          }

          if ((fabs(m_vecW[nm]) + anorm) == anorm) {
            break;
          }
	}

        if (flag) {
          c=0.0;
          s=1.0;
          for (i=l; i<k+1; i++) {
            f=s*rv1[i];
            rv1[i]=c*rv1[i];
            if ((fabs(f) + anorm) == anorm) {
              break;
            }
            g=m_vecW[i];
            h=pythag(f,g);
            m_vecW[i]=h;
            h=1.0/h;
            c=g*h;
            s = -f*h;
            for (j=0; j<m_dimm; j++) {
              y=m_matA[j][nm];
              z=m_matA[j][i];
              m_matA[j][nm]=y*c+z*s;
              m_matA[j][i]=z*c-y*s;
            }
          }
        }

        z=m_vecW[k];

        if (l == k) {
          if (z < 0.0) {
            m_vecW[k] = -z;
            for (j=0; j<m_dimn; j++) {
              m_matV[j][k] = -m_matV[j][k];
            }
          }
          break;
        }

        if (its == 29) {
          std::stringstream ss (std::stringstream::in | 
				std::stringstream::out);
	  ss << "No convergence in 30 iterations";
	  throw tExceptionObject(__FILE__,__LINE__,ss.str());
        }

        x=m_vecW[l];
        nm=k-1;
        y=m_vecW[nm];
        g=rv1[nm];
        h=rv1[k];
        f=((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
        g=pythag(f,1.0);
        f=((x-z)*(x+z)+h*((y/(f+sign(g,f)))-h))/x;
        c=s=1.0;
        for (j=l; j<=nm; j++) {
          i=j+1;
          g=rv1[i];
          y=m_vecW[i];
          h=s*g;
          g=c*g;
          z=pythag(f,h);
          rv1[j]=z;
          c=f/z;
          s=h/z;
          f=x*c+g*s;
          g=g*c-x*s;
          h=y*s;
          y *= c;
          for (jj=0; jj<m_dimn; jj++) {
            x=m_matV[jj][j];
            z=m_matV[jj][i];
            m_matV[jj][j]=x*c+z*s;
            m_matV[jj][i]=z*c-x*s;
          }

          z=pythag(f,h);

          m_vecW[j]=z;
          if (z) {
            z=1.0/z;
            c=f*z;
            s=h*z;
	  }

          f=c*g+s*y;
          x=c*y-s*g;
          for (jj=0; jj<m_dimm; jj++) {
            y=m_matA[jj][j];
            z=m_matA[jj][i];
            m_matA[jj][j]=y*c+z*s;
            m_matA[jj][i]=z*c-y*s;
          }
        }

        rv1[l]=0.0;
        rv1[k]=f;
        m_vecW[k]=x;
      }
    }

    // release memory allocated for "rv1"
    delete[] rv1;
  }


  // -----------------------------------------------------------------
  // Method: bksb()
  // 
  // Computes  the  solution of  the  linear  system AX  -  B  = 0  by
  // backsubstitution from the SVD of  matrix A. This method is a copy
  // of the procedure "svbksb" from the book Numerical Recipes in C++:
  // The Art of Scientific Computing,  by W.  Press, S.  Teukolsky, W.
  // Vetterling,   and  B.    Flannery,  Secoond   Edition,  Cambridge
  // University Press, 2002.
  //
  // -----------------------------------------------------------------
  void
  tSVDdcmp::bksb(
	         double* b,
	         double* x
                )
    const
  {
    int jj,j,i;
    double s;

    double* tmp = (double *) new double[m_dimn];

    for (j=0; j<m_dimn; j++) {
      s=0.0;
      if (m_vecW[j] != 0.0) {
        for (i=0; i<m_dimm; i++) {
          s += m_matA[i][j]*b[i];
        }

        s /= m_vecW[j];
      }

      tmp[j]=s;
    }

    for (j=0; j<m_dimn; j++) {
      s=0.0;
      for (jj=0; jj<m_dimn; jj++) {
        s += m_matV[j][jj]*tmp[jj];
      }
      x[j]=s;
    }

    // release memory allocated for "tmp"
    delete[] tmp;
  }


  // -----------------------------------------------------------------
  // Method: pythag()
  // 
  // Auxiliary  method  used  by  "SVDecomp".  It provides  a  way  of
  // robustly computing  the square root of  the sum of  the square of
  // two real numbers.
  //
  // -----------------------------------------------------------------
  double
  tSVDdcmp::pythag(
                   double a,
                   double b
		  )
    const
  {
    double absa, absb;

    absa=fabs(a);
    absb=fabs(b);
    if (absa > absb) {
      return absa*sqrt(1.0+SQR(absb/absa));
    }

    return (absb == 0.0) ? 0.0 : absb*sqrt(1.0+SQR(absa/absb));
  }


  // -----------------------------------------------------------------
  // Method: sign()
  // 
  // Returns the magnitude of a given number times the sign of another
  // given number.
  //
  // -----------------------------------------------------------------
  double
  tSVDdcmp::sign(const double& a, const double& b) const
  {
    return b >= 0 ? (a >= 0 ? a : -a) : (a >= 0 ? -a : a);
  }


}        // end of namespace common
