/** 
 * @class dcel::tComplex
 *  
 * @brief A  class implementing a two-dimensional cell  complex in the
 * DCEL data structure.
 *
 * @author
 * Marcelo Ferreira Siqueira @n
 * Universidade Federal de Mato Grosso do Sul, @n
 * Departamento de Computacao e Estatistica, @n
 * marcelo at dct (dot) ufms (dot) br
 *
 * @version 1.0
 * @date Abril 2008
 *
 * @attention This program is distributed  WITHOUT ANY WARRANTY,  and
 *            it may be freely redistributed under the condition that
 *            the copyright notices are  not  removed,  and  no
 *            compensation  is  received. Private, research, and
 *            institutional use  is free. Distribution of this code as
 *            part of a commercial system is permissible ONLY BY DIRECT
 *            ARRANGEMENT WITH THE AUTHOR.
 */

#include "complex.h"

#include <iostream>
#include <sstream>

#include <stdlib.h>

#include "macros.h"

#include "dcel.h"
#include "point.h"


namespace dcel {


  /** 
   * Builds a cell complex from a set of vertices and faces.
   *
   * @param nverts: the number of vertices.
   * @param lverts: an array with the vertex coordinates.
   * @param nfaces: the number of facets.
   * @param lfaces: an array with the facet identifiers.
   * @param nvpf: the number of  vertices per face.
   *
   */
  void
  tComplex::build(
                  unsigned nverts,
                  double* lverts,
                  unsigned nfaces,
                  unsigned* lfaces,
                  unsigned nvpf
                 )
    throw (tExceptionObject) 
  {
    //
    // Create the vertices of the complex.
    //

    /** Define a hash table to keep track of vertex IDs. */
    VMAP vmap;

    /** Create the vertices and initialize the table. */
    create_vertices(nverts,lverts,vmap);

    //
    // Create the faces and halfedges of the complex.
    //

    /** Define a hash table to keep track of the halfedge IDs. */
    HMAP hmap;

    /** Create the faces and halfedges and initialize the table. */
    try {
      create_faces(nfaces,lfaces,nvpf,hmap,vmap);
    }
    catch (const tExceptionObject& xpt) {
      treatException(xpt);
      exit(0);
    }

    /** Release memory associated with the vertex ID table. */
    vmap.clear();

    //
    // Create the edges of the complex.
    //
    create_edges(hmap);

    /** Release memory associated with the halfedge ID table. */
    hmap.clear();

    /** Check  if  the  cell   complex  just  built  is  topologically
     *  consistent. 
     */
    if (!is_consistent()) {
      std::stringstream ss(std::stringstream::in | 
                           std::stringstream::out);
      ss << "Input cell complex is topologically inconsistent.";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }
  }


  /** 
   * Creates the vertices of this cell complex.
   *
   * @param nverts: the number of vertices.
   * @param lverts: an array with the vertex coordinates.
   * @param  vmap: a table  of pairs  of the  form (vertex  ID, vertex
   * pointer).
   *
   */
  void
  tComplex::create_vertices(
                            unsigned nverts,
                            double* lverts,
                            VMAP& vmap
                           )
  {
    unsigned i;
    for (i=0; i<nverts; i++) {
      const unsigned j = 3 * i;

      double x = lverts[j  ];
      double y = lverts[j+1];
      double z = lverts[j+2];

      vtPointer vp = tVertex::NEW(
                                  tPoint(x,y,z),
                                  hePointer(0),
                                  tVertex::atPointer(0)
                                 );

      vmap.insert(std::make_pair(i,vp));
    }
  }


  /** 
   * Creates the faces of a cell complex.
   *
   * @param nfaces: the number of facets.
   * @param lfaces: an array with the facet identifiers.
   * @param nvpf: the number of  vertices per face (this is a constant
   * number).
   * @param  hmap:  a  table  of  pairs of  the  form  (half-edge  ID,
   * half-edge pointer).
   * @param  vmap: a table  of pairs  of the  form (vertex  ID, vertex
   * pointer).
   *
   */
  void
  tComplex::create_faces(
                         unsigned nfaces,
                         unsigned* lfaces,
                         unsigned nvpf,
                         HMAP& hmap,
                         const VMAP& vmap
                        )
    throw (tExceptionObject) 
  {
    /**
     * Verify if the number of vertices per face is at least 3.
     */
    if (nvpf < 3) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "Number of vertex per face must be at least 3.";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    /**
     * Loop  over the array  of edge  ID's and  for each  iteration i,
     * create the i-th facet of the complex.
     *
     */

    unsigned i;
    unsigned j = 0;
    for (i=0; i<nfaces; i++) {

      fcPointer fp = tFace::NEW(
                                hePointer(0),
                                Pointer(this),
                                tFace::atPointer(0)
                               );

      //
      // Create the halfedges of this facet.
      //
      hePointer *hv = new hePointer[nvpf];

      unsigned k;
      for (k=0; k<nvpf; k++) {
        /** 
         * Find the ID of the k-th vertex of the facet. 
         */
        VMAP::const_iterator vit = vmap.find(lfaces[j+k]);

        /** 
         * If k-th vertex cannot be found, throw an exception. 
         */
        if (vit == vmap.end()) {
          std::stringstream ss (std::stringstream::in | 
                                std::stringstream::out);
          ss << "Bad vertex ID.";
          throw tExceptionObject(__FILE__,__LINE__,ss.str());
        }

        vtPointer vp = vit->second;

        /** 
         * Allocate memory for the k-th halfedge. 
         */
        hv[k] = tHalfEdge::NEW(
                               vp,
                               egPointer(0),
                               fp,
                               hePointer(0),
                               hePointer(0),
                               tHalfEdge::atPointer(0)
                              );

        /** 
         * Set the halfedge pointer of the vertex pointed by "vp".
         */
        vp->set_hedge(hv[k]);
      }

      /** 
       * Update the pointers to previous and next halfedges. 
       */
      hv[     0]->set_prev(hv[nvpf-1]);
      hv[nvpf-1]->set_next(hv[     0]);

      for (k=1; k<nvpf; k++) {
        hv[k-1]->set_next(hv[k  ]);
        hv[k  ]->set_prev(hv[k-1]);
      }

      /** 
       * Set the pointer to the first halfedge of the face. 
       */
      fp->set_hedge(hv[0]);

      /** 
       * Add information to the table of halfedge IDs. 
       */
      for (k=0; k<nvpf; k++) {
        const unsigned v1 = lfaces[j+k];
        const unsigned v2 = lfaces[j+((k+1) % nvpf)];

        hmap.insert(std::make_pair(std::make_pair(v1,v2),hv[k]));
      }

      /** 
       * Release memory for the "hv" vector. 
       */
      delete[] hv;

      /** 
       * Update the facet vertex counter. 
       */
      j += nvpf;

      /** 
       * Insert facet into the list of facets of this complex. 
       */
      m_lfaces.push_back(fp);
    }


    /** 
     * Place in  the list of vertices  only the vertices  that are the
     * origin  of some  half-edge. The  ones  that are  not should  be
     * ignored.
     */
    VMAP::const_iterator vit;
    for (vit = vmap.begin(); vit != vmap.end(); ++vit) {
      if (vit->second->get_hedge().is_not_null()) {
        m_lverts.push_back(vit->second);
      }
    }
  }


  /** 
   * Creates the edges of this cell complex.
   *
   * @param hmap:  a hash table that  stores pairs of  halfedge ID and
   * pointer to halfedge.
   *
   */
  void
  tComplex::create_edges(HMAP& hmap)
  {
    /**
     * Loop over  the set of half-edges  of the cell  complex and, for
     * each  of them,  create  an edge  (if  the edge  does not  exist
     * already)  and update  the edge  and half-edge  pointers  of the
     * half-edges and edges, respectively.
     *
     */
    HMAP::iterator hit;
    for (hit = hmap.begin(); hit != hmap.end(); ++hit) {
      /**
       * If the  edge pointer  of the current  half-edge has  not been
       * assigned an edge yet, do  it now. Otherwise, skip to the next
       * loop iteration.
       */
      if (hit->second->get_edge().is_null()) {
        /**
         * Create a new edge. 
         */
        egPointer ep = tEdge::NEW(
                                  hit->second,
                                  hePointer(0),
                                  tEdge::atPointer(0)
                                 );

        /** 
         * Sets the edge pointer of the current half-edge. 
         */
        hit->second->set_edge(ep);

        /** 
         * Insert edge into the list of edges of this cell complex.
         */
        m_ledges.push_back(ep);

        /** 
         * Get the ID of the half-edge. 
         */
        unsigned v1 = hit->first.first;
        unsigned v2 = hit->first.second;

        /** 
         * Verify if  the current  half-edge has a  mate.  If  it does
         * not, throw an exception.
         */
        HMAP::iterator ait = hmap.find(std::make_pair(v2,v1));
        if (ait != hmap.end()) {
          ait->second->set_edge(ep);
          ep->set_hedge2(ait->second);
        }
      }
    }
  }


  /** 
   * Performs a topological consistency check in a cell complex.
   *
   * @return True if the cell  complex passed the consistency test and
   * false otherwise.
   *
   */
  bool
  tComplex::is_consistent() const
  {
    /** 
     * Check the halfedge and face pointers. 
     */
    fcIterator fit;
    for(fit = fc_begin(); fit != fc_end(); ++fit) {
      /** 
       * The face pointer to the complex cannot be null. 
       */
      if ((*fit)->get_complex().is_null()) {
        return false;
      }

      /** 
       * Get  the pointer  to  the  first half-edge  of  the cycle  of
       * half-edges of the face.
       */
      hePointer he1 = (*fit)->get_hedge();

      /** 
       * For each  half-edge of the  cycle of half-edges of  the face,
       * check the pointers.
       */
      hePointer he2 = he1;
      do {
        /** 
         * The pointer to the half-edge cannot be null. 
         */
        if (he2.is_null()) {
          return false;
        }

        /** 
         * The half-edge pointer to the  face must agree with the face
         * pointer.
         */
        else if (*fit != he2->get_face()) {
          return false;
        }

        /** 
         * The half-edge pointer to the next half-edge in the cycle of
         * half-edges of the face cannot be null.
         */
        else if (he2->next().is_null()) {
          return false;
        }

        /** The  half-edge  pointer to  the  origin  vertex cannot  be
	 *  null. 
         */
        if (he2->get_vert().is_null()) {
          return false;
        }

        /** 
         * The pointer to the previous half-edge of the next half-edge
	 *  ("he3") cannot be null and  must agree with the pointer to
	 *  the half-edge ("he2") being checked currently.
         */
        hePointer he3 = he2->next();
        if (he3->prev().is_null()) {
          return false;
        }
        else if (he3->prev() != he2) {
          return false;
        }

        /** 
         * Get the  next half-edge  of the cycle  of half-edge  of the
         * current face.
         */
        he2 = he2->next();
      } 
      while (he2 != he1);


      do {
        /** 
         * The half-edge pointer to the edge cannot be null. 
         */
        if (he2->get_edge().is_null()) {
          return false;
        }

        /** 
         * The half-edge  pointer to the  edge must agree with  one of
         * the two pointers of the edge to one of its half-edge.
         */
        hePointer he3 = he2->get_edge()->get_hedge();
        if ((he2 != he3) && (he2 != MATE(he3))) {
          return false;
	}

        /** 
         * Get the mate of the current half-edge. 
         */
        he3 = MATE(he2);

        /** 
         * If  the mate  of  the current  half-edge  exists, then  its
         * pointer to  the edge  must agree with  that of  the current
         * half-edge.
         */
        if (he3.is_not_null() && (he2->get_edge() != he3->get_edge())) {
          return false;
        }

        /** 
         * If  the mate  of  the current  half-edge  exists, then  its
         * pointer  to the  mate must  agree with  the pointer  to the
         * current half-edge.
         */
        else if (he3.is_not_null() && (he2 != MATE(he3))) {
          return false;
        }

        /** 
         * Get the  next half-edge  of the cycle  of half-edge  of the
         * current face.
         */
        he2 = he2->next();
      } 
      while (he2 != he1);

    } // end of the for loop for checking faces.


    /**
     * Check the vertex pointers.
     *
     */

    vtIterator vit;
    for(vit = vt_begin(); vit != vt_end(); ++vit) {
      /** 
       * Get  the pointer  to one  of  the half-edges  of the  current
       * vertex.
       */ 
      hePointer he1 = (*vit)->get_hedge();

      /** 
       * If the pointer  to the half-edge is null,  then the vertex is
       * ignored.
       */
      if (he1.is_null()) {
        continue;
      }

      /** 
       * The half-edge  pointer to the  origin vertex must  agree with
       * the pointer to the current vertex.
       */
      if (he1->get_vert() != *vit) {
        return false;
      }

      /** 
       * Get the  next half-edge in a counterclowise  traversal of the
       * edges incident to the current vertex.
       */
      hePointer he2 = he1->prev();

      he2 = MATE(he2);

      bool bd = false;
      do {
        /** 
         * If the edge is on the boundary, then we end the traversal.
         */
        if (he2.is_null()) {
          bd = true;
        }
        else {
          /** 
           * The  half-edge pointer  to the  origin vertex  must agree
           * with the pointer to the current vertex.
           */
          if (he2->get_vert() != *vit) {
            return false;
          }
          else {
            he2 = he2->prev();
            he2 = MATE(he2);
          }
        }
      }
      while ((he2 != he1) && !bd);

      if (bd) {
        he2 = MATE(he1);
        bd = false;
        do {
          /** 
           * If  the  edge  is  on  the  boundary,  then  we  end  the
           * traversal.
           */
          if (he2.is_null()) {
            bd = true;
          }
          else {
            he2 = he2->next();

            /** 
             * The half-edge  pointer to the origin  vertex must agree
             * with the pointer to the current vertex.
             */
            if (he2->get_vert() != *vit) {
              return false;
            }
            else {
              he2 = MATE(he2);
            }
          }
        }
        while ((he2 != he1) && !bd);
      }

    } // end of the for loop for checking vertices


    /**
     * Check the edge pointers.
     */
    egIterator eit;
    for(eit = eg_begin(); eit != eg_end(); ++eit) {
      /** 
       * Get  the pointer  to the first half-edge of the edge.
       */ 
      hePointer he1 = (*eit)->get_hedge();

      /** 
       * The pointer to the first half-edge cannot be null.
       */
      if (he1.is_null()) {
        return false;
      }


      /**
       * Check the edge pointer of the first half-edge.
       */
      if (he1->get_edge() != *eit) {
        return false;
      }


      /** 
       * Get the pointer to the second half-edge of the edge.
       */ 
      hePointer he2 = MATE(he1);


      /**
       * Check the edge pointer of the second half-edge.
       */
      if (he2.is_not_null() && (he2->get_edge() != *eit)) {
        return false;
      }
    }

    return true;
  }


  /** 
   * Verifies whether the cell complex is a simplicial complex.
   *
   * @return  True if  the cell  complex is  a simplicial  complex and
   * false otherwise.
   *
   */
  bool
  tComplex::is_simplicial() const
  {
    fcIterator fit;
    for(fit = fc_begin(); fit != fc_end(); ++fit) {
      hePointer he1 = (*fit)->get_hedge();
      hePointer he2 = he1->next()->next()->next();
      if (he1 != he2) {
        return false;
      }
    }

    return true;
  }


  /** 
   * Check if the cell complex has a boundary. 
   *
   * @return  True  if the  cell  complex  has  a boundary  and  false
   * otherwise.
   *
   */
  bool
  tComplex::has_boundary() const
  {
    egIterator eit;
    for (eit = eg_begin(); eit != eg_end(); ++eit) {
      if ((*eit)->is_on_boundary()) {
        return true;
      }
    }

    return false;
  }


  /** 
   * Removes all  smart pointer references  in the structure,  so that
   * all circular dependencies are  removes and memory can be released
   * correctly.
   *
   */
  void
  tComplex::cleanup()
  {
    std::list<vtPointer>::iterator vit;

    for (vit = m_lverts.begin(); vit != m_lverts.end(); ++vit) {
      (*vit)->set_hedge(hePointer(0));
 
      vaPointer va = (*vit)->get_attrib();

      if (va.is_not_null()) {
        va->set_owner(vtPointer(0));
        (*vit)->set_attrib(vaPointer(0)); 
      }
    }


    std::list<egPointer>::iterator eit;

    for (eit = m_ledges.begin(); eit != m_ledges.end(); ++eit) {
      hePointer he1 = (*eit)->get_hedge();
      hePointer he2 = MATE(he1);

      eaPointer ea = (*eit)->get_attrib();

      if (ea.is_not_null()) {
        ea->set_owner(egPointer(0));
        (*eit)->set_attrib(eaPointer(0));
      }

      (*eit)->set_hedge1(hePointer(0));
      (*eit)->set_hedge2(hePointer(0));
  
      haPointer ha = he1->get_attrib();

      if (ha.is_not_null()) {
        ha->set_owner(hePointer(0));
        he1->set_attrib(haPointer(0));
      }

      he1->set_vert(vtPointer(0));
      he1->set_edge(egPointer(0));
      he1->set_face(fcPointer(0));
      he1->set_next(hePointer(0));
      he1->set_prev(hePointer(0));

      if (he2.is_not_null()) {
        ha = he2->get_attrib();

        if (ha.is_not_null()) {
          ha->set_owner(hePointer(0));
          he2->set_attrib(haPointer(0));
        }

        he2->set_vert(vtPointer(0));
        he2->set_edge(egPointer(0));
        he2->set_face(fcPointer(0));
        he2->set_next(hePointer(0));
        he2->set_prev(hePointer(0));
      }
    }    

    std::list<fcPointer>::iterator fit;

    for (fit = m_lfaces.begin(); fit != m_lfaces.end(); ++fit) {
      (*fit)->set_hedge(hePointer(0));
      (*fit)->set_complex(Pointer(0));
 
      faPointer fa = (*fit)->get_attrib();

      if (fa.is_not_null()) {
        fa->set_owner(fcPointer(0));
        (*fit)->set_attrib(faPointer(0)); 
      }
    }


    atPointer ca = this->get_attrib();

    if (ca.is_not_null()) {
      ca->set_owner(Pointer(0));
      this->set_attrib(atPointer(0)); 
    }

    m_lverts.clear();
    m_ledges.clear();
    m_lfaces.clear();

    return;
  }


}
