/** 
 * @class dcel::tEdge
 *  
 * @brief A class implementing an edge of the DCEL data structure.
 *
 * @author
 * Marcelo Ferreira Siqueira @n
 * Universidade Federal de Mato Grosso do Sul, @n
 * Departamento de Computacao e Estatistica, @n
 * marcelo at dct (dot) ufms (dot) br
 *
 * @version 1.0
 * @date Abril 2008
 *
 * @attention This program is distributed  WITHOUT ANY WARRANTY,  and
 *            it may be freely redistributed under the condition that
 *            the copyright notices are  not  removed,  and  no
 *            compensation  is  received. Private, research, and
 *            institutional use  is free. Distribution of this code as
 *            part of a commercial system is permissible ONLY BY DIRECT
 *            ARRANGEMENT WITH THE AUTHOR.
 */

#ifndef EDGE_H_
#define EDGE_H_

#include "smartptr.h"
#include "dataobj.h"
#include "egattrib.h"
#include "halfedge.h"


namespace dcel {


  /**
   * Forward definition of classes.
   */
  class tComplex;


  /**
   * Class tEdge: implements an edge of the DCEL data structure.
   *
   */ 
  class tEdge : public tDataObject {
  public:
    /**
     * Defines Self as this class
     */    
    typedef tEdge                 Self;


    /**
     * Defines Pointer as smart pointer to this class.
     */    
    typedef tSmartPtr<Self>       Pointer;


    /**
     * Defines ConstPointer as a const smart pointer to this class.
     */    
    typedef tSmartPtr<const Self> ConstPointer;


    /**
     * Defines hePointer as smart pointer to tHalfEdge.
     */    
    typedef tHalfEdge::Pointer    hePointer;


    /**
     * Defines atPointer as a smart pointer to a tEGAttrib object.
     */
    typedef tSmartPtr<tEGAttrib>  atPointer;
     

    /**
     * Method for creating an instance of this object.
     */
    static Pointer NEW(
                       const hePointer& h1,
                       const hePointer& h2,
                       const atPointer& at
                      )
    {
      tEdge *dumbPtr = new tEdge;
      Pointer smartPtr(dumbPtr);

      smartPtr->set_hedge1(h1);
      smartPtr->set_hedge2(h2);
      smartPtr->set_attrib(at);

      return smartPtr;
    }


    /**
     * Destructor.
     */
    ~tEdge() {}


    /**
     * This method  returns a pointer  to the first half-edge  of this
     * edge.
     *
     * @return A pointer to a half-edge.
     *
     */    
    hePointer get_hedge() const
      { return m_he1; }


    /**
     * This method returns  a pointer to the second  half-edge of this
     * edge.
     *
     * @return A pointer to a half-edge.
     *
     */    
    hePointer get_mate(const hePointer& he) const
      {
	if (he == m_he1) return m_he2;

        return m_he1;
      }


    /**
     * Returns a pointer to the set of attributes of this edge.
     *
     * @return A pointer to a set of edge attributes.
     *
     */
    atPointer get_attrib() const
      { return m_at; }


    /**
     * Assigns a given pointer to the pointer to the set of attributes
     * of this edge.
     *
     * @param at: a pointer to a set of edge attributes.
     *
     */
    void set_attrib(const atPointer& at)
      { m_at = at; }


    /**
     * This method returns  true if this edge belongs  to the boundary
     * of the cell complex and false otherwise.
     *
     * @return True if  this edge belongs to the  boundary of the cell
     * complex and false otherwise.
     */    
    bool is_on_boundary() const
      { return m_he2.is_null(); }


    /**
     * Makes tComplex a friend class.
     *
     */
    friend class tComplex;


  private:
    /**
     * Default constructor.
     */
    tEdge() {}


    /**
     * Assigns a given  pointer to the pointer to  the first half-edge
     * of this edge.
     *
     * @param he: a pointer to a half-edge.
     *
     */
    void set_hedge1(const hePointer& he)
      { m_he1 = he; }


    /**
     * Assigns a given pointer to  the pointer to the second half-edge
     * of this edge.
     *
     * @param he: a pointer to a half-edge.
     *
     */
    void set_hedge2(const hePointer& he)
      { m_he2 = he; }


    /**
     * Pointer to the first half-edge of this edge.
     */
    hePointer m_he1;


    /**
     * Pointer to the second half-edge of this edge.
     */
    hePointer m_he2;


    /**
     * Pointer to the set of attributes of this edge.
     */
    atPointer m_at;

  };


}


#endif   // EDGE_H_
