/** 
 * @class dcel::tFace
 *  
 * @brief A class implementing a face of the DCEL data structure.
 *
 * @author
 * Marcelo Ferreira Siqueira @n
 * Universidade Federal de Mato Grosso do Sul, @n
 * Departamento de Computacao e Estatistica, @n
 * marcelo at dct (dot) ufms (dot) br
 *
 * @version 1.0
 * @date Abril 2008
 *
 * @attention This program is distributed  WITHOUT ANY WARRANTY,  and
 *            it may be freely redistributed under the condition that
 *            the copyright notices are  not  removed,  and  no
 *            compensation  is  received. Private, research, and
 *            institutional use  is free. Distribution of this code as
 *            part of a commercial system is permissible ONLY BY DIRECT
 *            ARRANGEMENT WITH THE AUTHOR.
 */

#ifndef FACE_H_
#define FACE_H_

#include "smartptr.h"
#include "dataobj.h"
#include "fcattrib.h"
#include "halfedge.h"
#include "edge.h"


namespace dcel {


  /**
   * Forward definition of classes.
   */
  class tComplex;


  /**
   * Class tFace: implements a face of the DCEL data structure.
   *
   */ 
  class tFace : public tDataObject {
  public:
    /**
     * Defines Self as this class.
     */   
    typedef tFace                 Self;


    /**
     * Defines Pointer as smart pointer to this class.
     */ 
    typedef tSmartPtr<Self>       Pointer;


    /**
     * Defines Pointer as a const smart pointer to this class.
     */ 
    typedef tSmartPtr<const Self> ConstPointer;

    /**
     * Defines cpPointer as smart pointer to tComplex.
     */  
    typedef tSmartPtr<tComplex>   cpPointer;
 

    /** 
     * Defines hePointer as smart pointer to tHalfEdge.
     */ 
    typedef tHalfEdge::Pointer    hePointer;


    /** 
     * Defines egPointer as smart pointer to tEdge.
     */
    typedef tEdge::Pointer        egPointer;


    /**
     * Defines atPointer as a smart pointer to a tFCAttrib object.
     */
    typedef tSmartPtr<tFCAttrib>  atPointer;


    /**
     * Method for creating an instance of this object.
     *
     * @param h: a pointer to a half-edge.
     * @param c: a pointer to a cell complex.
     * @param a: a pointer to a set of facet attributes.
     *
     * @return  A pointer to  a newly  created facet.
     *
     */
    static Pointer NEW(
                       const hePointer& h,
                       const cpPointer& c,
                       const atPointer& a
                      )
    {
      tFace *dumbPtr = new tFace;
      Pointer smartPtr(dumbPtr);

      smartPtr->set_hedge(h);
      smartPtr->set_complex(c);
      smartPtr->set_attrib(a);

      return smartPtr;
    }


    /**
     * Virtual Destructor.
     */
    ~tFace() {}


    /**
     * Returns  a pointer  to  the  first half-edge  of  the cycle  of
     * half-edges of this face.
     *
     * @return A pointer to a half-edge.
     *
     */
    hePointer get_hedge() const
      { return m_he; }


    /**
     * Returns the cell complex containing this face.
     *
     * @return A pointer to a cell complex.
     *
     */
    cpPointer get_complex() const
      { return m_cp; }


    /**
     * Returns a pointer to the set of attributes of this face.
     *
     * @return A pointer to a set of face attributes.
     *
     */
    atPointer get_attrib() const
      { return m_at; }


    /**
     * Assigns a given pointer to the pointer to the set of attributes
     * of this face.
     *
     * @param at: a pointer to a set of face attributes.
     *
     */
    void set_attrib(const atPointer& at)
      { m_at = at; }


    /**
     * Makes tComplex a friend class.
     *
     */
    friend class tComplex;


  private:
    /**
     * Default constructor.
     */
    tFace() {}


    /**
     * Assigns a given  pointer to the pointer to  the first half-edge
     * of this face.
     *
     * @param he: a pointer to a half-edge.
     *
     */
    void set_hedge(const hePointer& he)
      { m_he = he; }


    /** 
     * Assigns a given pointer to the pointer to the cell complex this
     * face belongs to.
     *
     *
     * @param cp: a pointer to the cell complex this face belongs to.
     *
     */
    void set_complex(const cpPointer& cp)
      { m_cp = cp; }


    /** 
     * A pointer to the first  half-edge of the cycle of half-edges of
     * this face.
     *
     */
    hePointer m_he;


    /**
     * A pointer to the cell complex this face belongs to.
     */
     cpPointer m_cp;


    /**
     * A pointer to the set of attributes of this face.
     */
    atPointer m_at;

  };


}

#endif   // FACE_H_

