/** 
 * @class dcel::tVertex
 *  
 * @brief A class implementing a vertex of the DCEL data structure.
 *
 * @author
 * Marcelo Ferreira Siqueira @n
 * Universidade Federal de Mato Grosso do Sul, @n
 * Departamento de Computacao e Estatistica, @n
 * marcelo at dct (dot) ufms (dot) br
 *
 * @version 1.0
 * @date Abril 2008
 *
 * @attention This program is distributed  WITHOUT ANY WARRANTY,  and
 *            it may be freely redistributed under the condition that
 *            the copyright notices are  not  removed,  and  no
 *            compensation  is  received. Private, research, and
 *            institutional use  is free. Distribution of this code as
 *            part of a commercial system is permissible ONLY BY DIRECT
 *            ARRANGEMENT WITH THE AUTHOR.
 */

#ifndef VERTEX_H_
#define VERTEX_H_

#include "smartptr.h"
#include "dataobj.h"
#include "vtattrib.h"
#include "point.h"


namespace dcel {


  /**
   * Forward definition of classes.
   */
  struct tComplex;
  struct tHalfEdge;

  /**
   * Class tVertex: implements a vertex of the DCEL data structure.
   *
   */ 
  class tVertex : public tDataObject {
  public:
    /**
     * Defines Self as this class.
     */
    typedef tVertex               Self;


    /**
     * Defines Pointer as a smart pointer to this class.
     */
    typedef tSmartPtr<Self>       Pointer;

    
    /**
     * Defines ConstPointer as a const smart pointer to this class.
     */
    typedef tSmartPtr<const Self> ConstPointer;


    /**
     * Defines hePointer as a smart pointer to a tHalfEdge object.
     */
    typedef tSmartPtr<tHalfEdge>  hePointer;


    /**
     * Defines atPointer as a smart pointer to a tVTAttrib object.
     */
    typedef tSmartPtr<tVTAttrib>  atPointer;


    /**
     * Method for creating an instance of this object.
     *
     * @param p: a 3D point.
     * @param h: a pointer to a half-edge.
     * @param a: a pointer to a set of vertex attributes.
     *
     * @return A pointer to a newly created vertex.
     *
     */
    static Pointer NEW(
                       const tPoint& p,
                       const hePointer& h,
                       const atPointer& a
                      )
    {
      tVertex *dumbPtr = new tVertex;
      Pointer smartPtr(dumbPtr);

      smartPtr->set_point(p);
      smartPtr->set_hedge(h);
      smartPtr->set_attrib(a);

      return smartPtr;
    }


    /**
     * Virtual Destructor.
     */
    virtual ~tVertex() {}


    /**
     * Returns the location of this vertex.
     *
     * @return A 3D point.
     */
    tPoint get_point() const
      { return m_pt; }


    /**
     * Returns a  pointer to a  half-edge whose origin vertex  is this
     * vertex.
     *
     * @return A pointer to a half-edge.
     */
    hePointer get_hedge() const
      { return m_he; }


    /**
     * Returns a pointer to the attributes of this vertex.
     *
     * @return A pointer to a set of vertex attributes.
     */
    atPointer get_attrib() const
      { return m_at; }


    /**
     * Assigns a given pointer to the pointer to the set of attributes
     * of this vertex.
     *
     * @param at: a pointer to a set of vertex attributes.
     */
    void set_attrib(const atPointer& at)
      { m_at = at; }


    /**
     * Makes tComplex a friend class.
     *
     */
    friend class tComplex;


  private:
    /**
     * Default constructor.
     */
    tVertex() {}


    /**
     * Assigns a location to this vertex.
     *
     * @param p: a 3D point.
     */
    void set_point(const tPoint& p)
      { m_pt = p; }


    /**
     * Assigns a given pointer to  the pointer to some half-edge whose
     * origin is this vertex.
     *
     * @param he: a pointer to a half-edge.
     *
     */
    void set_hedge(const hePointer& he)
      { m_he = he; }


    /**
     * Vertex location.
     */
    tPoint m_pt;


    /**
     * A  pointer  to some  half-edge  whose  initial  vertex is  this
     * vertex.
     *
     */
    hePointer m_he;


    /**
     * A pointer to the set of attributes of this vertex.
     */
    atPointer m_at;

  };

}

#endif  // VERTEX_H_
