//  ===================================================================
//
//  Program:  OFFLIB
//  Module:   reader.cpp
//  Language: C++
//  Date:     February 2008
//  Version:  1.0.0
//
//  Copyright (c) 2008
//
//  Marcelo Siqueira
//
//  Departamento de Computao e Estatstica
//  Universidade Federal de Mato Grosso do Sul
//  Av. Costa e Silva, S/N, Campus Universitria
//  Campo Grande, MS, Brasil, 79090-700
//  Phone No.: +55 67 3345 7458
//             +55 67 3345 7515
//
//  marcelo@dct.ufms.br
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
//  ===================================================================

#include "reader.h"

#include <iostream>
#include <sstream>
#include <stdlib.h>

#include "macros.h"

namespace offlib {


  using std::cout;
  using std::endl;


  // -----------------------------------------------------------------
  //
  // Method: read()
  // 
  // Reads the  vertices and  faces of a  triangular mesh from  an OFF
  // file.
  //
  // -----------------------------------------------------------------
  void
  tReader::read(
                unsigned& nv,
                double*& vset,
                unsigned& nf,
                unsigned*& fset
	       )
    throw (tExceptionObject) 
  {
    std::filebuf fb;
    if (fb.open(m_fname.c_str(), std::ios::in) == 0) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "Unable to open input file " << m_fname;
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    std::istream istr(&fb);

    m_is = new tLexer(istr);

    vset = 0;
    fset = 0;

    try {
      read_header(nv,nf);
      read_verts(nv,vset);
      read_faces(nf,nv,fset);
    }
    catch (const tExceptionObject& xpt) {
      treatException(xpt);
      fb.close();
      if (vset != 0) {
        delete[] vset;
        vset = 0;
      }
      if (fset != 0) {
        delete[] fset;
        fset = 0;
      }
      exit(0);
    }

    fb.close();
  }


  // -----------------------------------------------------------------
  // Method: read_header()
  // 
  // Reads the number of vertices  and faces of a triangular mesh from
  // an OFF file.
  //
  // -----------------------------------------------------------------
  void
  tReader::read_header(
                       unsigned& nv,
                       unsigned& nf
                      )  
    throw (tExceptionObject) 
  {

    //
    // The first line should start with the string "OFF"
    //
    std::string stroff;
    if (!m_is->get_str(stroff)) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "Error while trying to string at position "
         << m_is->charcount()
         << " of "
         << m_is->linecount()
         << " of file "
         << m_fname.c_str(); 
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if (stroff != "OFF") {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << " The first line should start with the string OFF";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }
   
    //
    // Read in the number of vertices of the mesh
    //
    if (!m_is->get_uint(nv)) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "Error while trying to read integer at position "
         << m_is->charcount()
         << " of "
         << m_is->linecount()
         << " of file "
         << m_fname.c_str(); 
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if (nv < 4) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "Number of vertices must be at least 4";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    //
    // Read in the number of faces of the mesh
    //
    if (!m_is->get_uint(nf)) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "Error while trying to read integer at position "
         << m_is->charcount()
         << " of "
         << m_is->linecount()
         << " of file "
         << m_fname.c_str(); 
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if (nf < 4) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "Number of faces must be at least 4";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    //
    // Read in the number of edges of the surface
    //
    unsigned ne;
    if (!m_is->get_uint(ne)) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "Error while trying to read integer at position "
         << m_is->charcount()
         << " of "
         << m_is->linecount()
         << " of file "
         << m_fname.c_str(); 
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    // There is no use for the integer "ne" here.
  }


  // -----------------------------------------------------------------
  // Method: read_verts()
  // 
  // Reads the vertices of a triangular mesh from an OFF file.
  //
  // -----------------------------------------------------------------
  void
  tReader::read_verts(
                      unsigned nv,
                      double*& vset
                     ) 
    throw (tExceptionObject) 
  {
    //
    // Allocate memory for storing the vertices.
    //
    vset = new double[3 * nv];

    //
    // Read "nv" lines with the vertex coordinates
    // 
    unsigned i = 0;
    while (i < nv) {
      double x;
      if (!m_is->get_double(x)) {
        std::stringstream ss (std::stringstream::in | 
                              std::stringstream::out);
        ss << "Error while trying to read float at position "
           << m_is->charcount()
           << " of "
           << m_is->linecount()
           << " of file "
           << m_fname.c_str(); 
        throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }

      double y;
      if (!m_is->get_double(y)) {
        std::stringstream ss (std::stringstream::in | 
                              std::stringstream::out);
        ss << "Error while trying to read float at position "
           << m_is->charcount()
           << " of "
           << m_is->linecount()
           << " of file "
           << m_fname.c_str(); 
        throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }

      double z;
      if (!m_is->get_double(z)) {
        std::stringstream ss (std::stringstream::in | 
                              std::stringstream::out);
        ss << "Error while trying to read float at position "
           << m_is->charcount()
           << " of "
           << m_is->linecount()
           << " of file "
           << m_fname.c_str(); 
        throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }

      const unsigned j = 3 * i;

      vset[j    ] = x;
      vset[j + 1] = y;
      vset[j + 2] = z;

      ++i;
    }
  }


  // -----------------------------------------------------------------
  // Method: read_faces()
  // 
  // Reads in the faces of a triangular mesh from an OFF file.
  //
  // -----------------------------------------------------------------
  void
  tReader::read_faces(
                      unsigned nf,
                      unsigned nv,
                      unsigned*& fset
                     ) 
    throw (tExceptionObject) 
  {
    //
    // Allocate memory for storing the vertices.
    //
    fset = new unsigned[3 * nf];

    //
    // Read "nf" lines of the form "3 v1 v2 v3", where "v1", "v2", and
    // "v3"  are the  vertex  indices  of the  face  defined by  these
    // vertices.
    //

    unsigned i = 0;
    while (i < nf) {
      unsigned nvaux;
      if (!m_is->get_uint(nvaux)) {
        std::stringstream ss (std::stringstream::in | 
                              std::stringstream::out);
        ss << "Error while trying to read integer at position "
           << m_is->charcount()
           << " of "
           << m_is->linecount()
           << " of file "
           << m_fname.c_str(); 
        throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }

      if (nvaux != 3) {
        std::stringstream ss (std::stringstream::in | 
                              std::stringstream::out);
        ss << "All faces must be triangles";
        throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }

      unsigned i1;
      if (!m_is->get_uint(i1)) {
        std::stringstream ss (std::stringstream::in | 
                              std::stringstream::out);
        ss << "Error while trying to read integer at position "
           << m_is->charcount()
           << " of "
           << m_is->linecount()
           << " of file "
           << m_fname.c_str(); 
        throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }

      if (i1 >= nv) {
        std::stringstream ss (std::stringstream::in | 
                              std::stringstream::out);
        ss << "Vertex index must be a number in the interval [0,"
           << nv - 1
           << "]";
        throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }

      unsigned i2;
      if (!m_is->get_uint(i2)) {
        std::stringstream ss (std::stringstream::in | 
                              std::stringstream::out);
        ss << "Error while trying to read integer at position "
           << m_is->charcount()
           << " of "
           << m_is->linecount()
           << " of file "
           << m_fname.c_str(); 
        throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }

      if (i2 >= nv) {
        std::stringstream ss (std::stringstream::in | 
                              std::stringstream::out);
        ss << "Vertex index must be a number in the interval [0,"
           << nv - 1
           << "]";
        throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }

      unsigned i3;
      if (!m_is->get_uint(i3)) {
        std::stringstream ss (std::stringstream::in | 
                              std::stringstream::out);
        ss << "Error while trying to read integer at position "
           << m_is->charcount()
           << " of "
           << m_is->linecount()
           << " of file "
           << m_fname.c_str(); 
        throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }

      if (i3 >= nv) {
        std::stringstream ss (std::stringstream::in | 
                              std::stringstream::out);
        ss << "Vertex index must be a number in the interval [0,"
           << nv - 1
           << "]";
        throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }

      // Store face data.
      const unsigned j = 3 * i;

      fset[j    ] = i1;
      fset[j + 1] = i2;
      fset[j + 2] = i3;

      ++i;
    }
  }


}        // end of namespace offlib
