/** 
 * @class dcel::tMyFCAttrib
 *  
 * @brief A class representing a set of half-edge attributes.
 *
 * @author
 * Marcelo Ferreira Siqueira @n
 * Universidade Federal de Mato Grosso do Sul, @n
 * Departamento de Computacao e Estatistica, @n
 * marcelo at dct (dot) ufms (dot) br
 *
 * @version 1.0
 * @date Abril 2008
 *
 * @attention This program is distributed  WITHOUT ANY WARRANTY,  and
 *            it may be freely redistributed under the condition that
 *            the copyright notices are  not  removed,  and  no
 *            compensation  is  received. Private, research, and
 *            institutional use  is free. Distribution of this code as
 *            part of a commercial system is permissible ONLY BY DIRECT
 *            ARRANGEMENT WITH THE AUTHOR.
 */


#ifndef MYFCATTRIB_H_
#define MYFCATTRIB_H_

#include "macros.h"
#include "exceptionobj.h"
#include "smartptr.h"
#include "dataobj.h"

#include "vertex.h"
#include "halfedge.h"
#include "edge.h"
#include "face.h"
#include "complex.h"
#include "fcattrib.h"


#define FCATTRIB(f) \
        tSmartPtr<tMyFCAttrib>(dynamic_cast<tMyFCAttrib*> \
        (f->get_attrib().get_pointee()))


namespace pps {


  class tMyFCAttrib : public dcel::tFCAttrib {
  public:
    /**
     * Defines Self as this class
     */    
    typedef tMyFCAttrib                Self;


    /**
     * Defines Pointer as smart pointer to this class.
     */    
    typedef tSmartPtr<Self>            Pointer;


    /**
     * Defines ConstPointer as a const smart pointer to this class.
     */    
    typedef tSmartPtr<const Self>      ConstPointer;


    /**
     * Defines fcPointer as a smart pointer to a tFace object.
     */
    typedef tSmartPtr<dcel::tFace>     fcPointer;


    /**
     * Type definition  for the  order of the  barycentric coordinates
     * associated with the affine frame corresponding to the face that
     * owns this attribute.
     */
    typedef enum { UVW=0 , VWU=1 , WUV=2 } BCORDER;


    /**
     * Method for creating an instance of this object.
     *
     * @param f: a pointer to a face.
     *
     */
    static Pointer NEW(
                       const fcPointer& f,
                       unsigned n,
                       BCORDER o,
                       double* cpts
                      ) 
    {
      tMyFCAttrib *dumbPtr = new tMyFCAttrib();
      Pointer smartPtr(dumbPtr);

      smartPtr->set_owner(f);

      smartPtr->m_n = n;
      smartPtr->m_o = o;
      smartPtr->m_c = cpts;

      return smartPtr;
    }


    /**
     * Destructor.
     */
    virtual ~tMyFCAttrib();


    /**
     * Returns the degree of the extraordinary vertex of the face that
     * owns this  attribute. If the face has  no extraordinary vertex,
     * it returns  the degree of the  first vertex of  the face, which
     * should always be equal to 6.
     *
     * @return The degree of the extraordinary vertex of the face that
     * owns this attribute.
     *
     */
     unsigned get_extra_vert_degree() const { return m_n; }


    /**
     * Returns  the order  of the  barycentric  coordinates associated
     * with the affine frame corresponding  to the face that owns this
     * attribute.
     *
     * @return  The order  of the  barycentric  coordinates associated
     * with the affine frame corresponding  to the face that owns this
     * attribute.
     *
     */
     BCORDER get_bary_coord_order() const { return m_o; }


    /**
     * Returns  the   i-th  projected  point  of   the  surface  patch
     * associated with the face that owns this attribute.
     */
     void get_point( unsigned , double* ) const 
       throw (tExceptionObject);

  private:
    /**
     * Default constructor.
     */
    tMyFCAttrib() {}


    /**
     * Linearizes the  index (i,j) of a bidimensional  matrix with "n"
     * columns.
     *
     */
    unsigned index(unsigned i, unsigned j, unsigned n) const
    {
      return j + ( i * n );
    }


    /**
     * Degree of  the only extraordinary  vertex (if any) of  the face
     * that  owns this  attribute. If  the face  has  no extraordinary
     * vertex, then the value of "m_n" must be 6.
     */
    unsigned m_n;


    /**
     * Order of the barycentric coordinates associated with the affine
     * frame corresponding to the face that owns this attribute.
     */
    BCORDER m_o;


    /**
     * A pointer  to array  of projected points  of the  surface patch
     * associated with the face that owns this attribute.
     */
    double* m_c;

  };


}


#endif   // MYFCATTRIB_H_
