/** 
 * @class dcel::tMyHEAttrib
 *  
 * @brief A class representing a set of half-edge attributes.
 *
 * @author
 * Marcelo Ferreira Siqueira @n
 * Universidade Federal de Mato Grosso do Sul, @n
 * Departamento de Computacao e Estatistica, @n
 * marcelo at dct (dot) ufms (dot) br
 *
 * @version 1.0
 * @date Abril 2008
 *
 * @attention This program is distributed  WITHOUT ANY WARRANTY,  and
 *            it may be freely redistributed under the condition that
 *            the copyright notices are  not  removed,  and  no
 *            compensation  is  received. Private, research, and
 *            institutional use  is free. Distribution of this code as
 *            part of a commercial system is permissible ONLY BY DIRECT
 *            ARRANGEMENT WITH THE AUTHOR.
 */

#include "myheattrib.h"

#include "dcel.h"

#include "mathmacros.h"

#include <cmath>
#include <cassert>


namespace pps {


  /**
   * Returns the  degree of  the origin vertex  of the  half-edge that
   * owns this attribute.
   *
   * @return The degree  of the origin vertex of  the half-edge that
   * owns this attribute.
   *
   */
  unsigned 
  tMyHEAttrib::get_degree() const
  {
    hePointer he = this->get_owner();

    assert( he.is_not_null() );

    vtPointer vt = he->get_vert();

    assert( vt.is_not_null() );

    vaPointer va = VTATTRIB( vt );

    assert( va.is_not_null() );

    return va->get_degree();
  } 


  /**
   * Computes  the rectangular  coordinates of  the image  of  a point
   * under the  transition map "g" associated with  the half-edge that
   * owns this attribute.
   *
   * @param x: X coordinate of the given point. 
   * @param y: Y coordinate of the given point.
   * @param r: X coordinate of the image of the given point under "g".
   * @param s: Y coordinate of the image of the given point under "g".
   *
   */
  void
  tMyHEAttrib::g(
                 double x,
                 double y,
                 double& u,
                 double& v
                ) 
    const 
  {
    const double MYPI = acos(-1);
 
    hePointer he1 = this->get_owner();

    assert(he1.is_not_null());

    hePointer he2 = MATE(he1);

    assert(he2.is_not_null());

    Pointer ha1 = HEATTRIB(he1);

    assert(ha1.is_not_null());

    Pointer ha2 = HEATTRIB(he2);

    assert(ha2.is_not_null());

    const unsigned nu =  ha1->get_degree();
    const unsigned nw =  ha2->get_degree();

    const double au = (2 * MYPI) / nu;
    const double aw = (2 * MYPI) / nw;

    const unsigned i = this->get_id();
    const unsigned j =  ha2->get_id();

    /**
     * Apply a rotation to move  the P-polygon edge onto the edge with
     * end vertices (0,0) and (1,0).
     */
    rot_2d(x,y,-(i * au),u,v);


    /**
     * It is  more convenient to  work with polar coordinates.  So, we
     * compute the "angle" and  the "length" associated with the point
     * "(u,v)".
     */
    double rr = sqrt( (u * u) + (v * v) );

    if (!ISZERO(rr)) {
      double aa;
      if (u < 0) {
        aa = (2 * MYPI) - acos( ABS(u) / rr );
      }
      else {
        aa = asin( v / rr );
      }

      /**
       * Map the point to the canonical quadrilateral.
       */

      aa *= ( nu / double(6) );
      rr *= ( cos( MYPI / 6 ) / cos( MYPI / nu ) );

      u = rr * cos(aa);
      v = rr * sin(aa);
    }

    /**
     * Compute the double reflection "H".
     */
    u = 1 - u;
    v =   - v;

    /**
     * Compute  the  polar coordinates  of  the  point  in the  target
     * P-domain.
     */
    rr = sqrt( (u * u) + (v * v) );

    if (!ISZERO(rr)) {
      double aa;
      if (u < 0) {
        aa = (2 * MYPI) - acos( ABS(u) / rr );
      }
      else {
        aa = asin( v / rr );
      }

      /**
       * Map the point to the target P-domain.
       */

      aa *= ( 6 / double(nw) );
      rr *= ( cos( MYPI / nw ) / cos( MYPI / 6 ) );

      u = rr * cos(aa);
      v = rr * sin(aa);
    }

    /**
     * Apply a rotation  to move the P-polygon edge  with end vertices
     * (0,0)  and (1,0)  onto the  edge  with end  vertices (0,0)  and
     * (cos(j*aw), sin(j*aw)).
     */
    rot_2d(u,v,(j * aw),u,v);

    if (ABS(u) < 1e-15) {
      u = 0;
    }

    if (ABS(v) < 1e-15) {
      v = 0;
    }

    if (ABS(1 - u) < 1e-15) {
      u = 1;
    }

    if (ABS(1 - v) < 1e-15) {
      v = 1;
    }
  }


  /**
   * Computes  the rectangular  coordinates of  the image  of  a point
   * under the map that takes the point to the canonical domain.  This
   * map is associated with the half-edge that owns this attribute.
   *
   * @param x: X coordinate of the given point. 
   * @param y: Y coordinate of the given point.
   * @param r: X coordinate of the image of the given point.
   * @param s: Y coordinate of the image of the given point.
   *
   */
  void
  tMyHEAttrib::to_cd(
                     double x,
                     double y,
                     double& u,
                     double& v
                    )  
    const 
  {
    const double MYPI = acos(-1);
 
    const unsigned nu = this->get_degree();

    const double au = (2 * MYPI) / nu;

    const unsigned i = this->get_id();

    /**
     * Apply a rotation to move  the P-polygon edge onto the edge with
     * end vertices (0,0) and (1,0).
     */
    rot_2d(x,y,-(i * au),u,v);


    /**
     * It is  more convenient to  work with polar coordinates.  So, we
     * compute the "angle" and  the "length" associated with the point
     * "(u,v)".
     */
    double rr = sqrt( (u * u) + (v * v) );

    if (!ISZERO(rr)) {
      double aa;
      if (u < 0) {
        aa = (2 * MYPI) - acos( ABS(u) / rr );
      }
      else {
        aa = asin( v / rr );
      }

      /**
       * Map the point to the canonical quadrilateral.
       */

      aa *= ( nu / double(6) );
      rr *= ( cos( MYPI / 6 ) / cos( MYPI / nu ) );

      u = rr * cos(aa);
      v = rr * sin(aa);
    }

    if (ABS(u) < 1e-15) {
      u = 0;
    }

    if (ABS(v) < 1e-15) {
      v = 0;
    }

    if (ABS(1 - u) < 1e-15) {
      u = 1;
    }

    if (ABS(1 - v) < 1e-15) {
      v = 1;
    }
  }


  /**
   * Computes  the rectangular  coordinates of  the image  of  a point
   * under the map  that takes the point from  the canonical domain to
   * the  domain   associated  with  the  half-edge   that  owns  this
   * attribute.
   *
   * @param x: X coordinate of the given point. 
   * @param y: Y coordinate of the given point.
   * @param r: X coordinate of the image of the given point.
   * @param s: Y coordinate of the image of the given point.
   *
   */
  void
  tMyHEAttrib::from_cd(
                       double x,
                       double y,
                       double& u,
                       double& v
                      ) 
    const 
  {
    const double MYPI = acos(-1);
 
    const unsigned nu = this->get_degree();

    const double au = (2 * MYPI) / nu;

    const unsigned i = this->get_id();

    /**
     * Compute  the  polar coordinates  of  the  point  in the  target
     * P-domain.
     */
    double rr = sqrt( (x * x) + (y * y) );

    if (!ISZERO(rr)) {
      double aa;
      if (x < 0) {
        aa = (2 * MYPI) - acos( ABS(x) / rr );
      }
      else {
        aa = asin( y / rr );
      }

      /**
       * Map the point to the target P-domain.
       */

      aa *= ( 6 / double(nu) );
      rr *= ( cos( MYPI / nu ) / cos( MYPI / 6 ) );

      u = rr * cos(aa);
      v = rr * sin(aa);
    }
    else {
      u = v = 0;
    }

    /**
     * Apply a rotation  to move the P-polygon edge  with end vertices
     * (0,0)  and (1,0)  onto the  edge  with end  vertices (0,0)  and
     * (cos(i*au), sin(u*au)).
     */
    rot_2d(u,v,(i * au),u,v);

    if (ABS(u) < 1e-15) {
      u = 0;
    }

    if (ABS(v) < 1e-15) {
      v = 0;
    }

    if (ABS(1 - u) < 1e-15) {
      u = 1;
    }

    if (ABS(1 - v) < 1e-15) {
      v = 1;
    }
  }


  /**
   * Computes the ID of the half-edge that owns this attribute.
   *
   */
  void
  tMyHEAttrib::compute_id() throw (tExceptionObject)
  {
    hePointer he1 = this->get_owner();

    assert( he1.is_not_null() );

    vtPointer vt = he1->get_vert();

    assert( vt.is_not_null() );

    hePointer he2 = vt->get_hedge();

    assert( he2.is_not_null() );

    unsigned id = 0;

    while (he1 != he2) {
      ++id;

      he2 = MATE( he2->prev() );

      assert( he2.is_not_null() );
    }

    this->set_id(id);
  }


  /**
   * Computes the  location of a  given point after a  rotation around
   * the origin by a given angle.
   *
   * @param x: X coordinate of the given point.
   * @param y: Y coordinate of the given point.
   * @param ang: a given rotation angle.
   * @param u: X coordinate of the point resulting  from the rotation
   * of (x,y) by ang.
   * @param v: Y  coordinate of the point resulting  from the rotation
   * of (x,y) by ang.
   *
   */
  void 
  tMyHEAttrib::rot_2d(
                      double  x,
                      double  y,
                      double  ang,
                      double& u,
                      double& v
                     )
    const
  {
    u = (x * cos(ang)) - (y * sin(ang));
    v = (x * sin(ang)) + (y * cos(ang));
  }

}
