/** 
 * @class dcel::tMyVTAttrib
 *  
 * @brief A class representing a set of vertex attributes.
 *
 * @author
 * Marcelo Ferreira Siqueira @n
 * Universidade Federal de Mato Grosso do Sul, @n
 * Departamento de Computacao e Estatistica, @n
 * marcelo at dct (dot) ufms (dot) br
 *
 * @version 1.0
 * @date Abril 2008
 *
 * @attention This program is distributed  WITHOUT ANY WARRANTY,  and
 *            it may be freely redistributed under the condition that
 *            the copyright notices are  not  removed,  and  no
 *            compensation  is  received. Private, research, and
 *            institutional use  is free. Distribution of this code as
 *            part of a commercial system is permissible ONLY BY DIRECT
 *            ARRANGEMENT WITH THE AUTHOR.
 */


#ifndef MYVTATTRIB_H_
#define MYVTATTRIB_H_

#include <sstream>

#include "exceptionobj.h"
#include "macros.h"
#include "smartptr.h"
#include "dataobj.h"

#include "vertex.h"
#include "halfedge.h"
#include "edge.h"
#include "face.h"
#include "complex.h"
#include "vtattrib.h"

#include "3dvector.h"
#include "bezier.h"


#define VTATTRIB(v) \
        tSmartPtr<tMyVTAttrib>(dynamic_cast<tMyVTAttrib*> \
        (v->get_attrib().get_pointee()))


namespace pps {


  using common::t3DVector;

  
  class tMyVTAttrib : public dcel::tVTAttrib {
  public:
    /**
     * Defines Self as this class
     */    
    typedef tMyVTAttrib                  Self;


    /**
     * Defines Pointer as smart pointer to this class.
     */    
    typedef tSmartPtr<Self>              Pointer;


    /**
     * Defines ConstPointer as a const smart pointer to this class.
     */    
    typedef tSmartPtr<const Self>        ConstPointer;


    /**
     * Defines vtPointer as a smart pointer to a tVertex object.
     */
    typedef tSmartPtr<dcel::tVertex>     vtPointer;


    /**
     * Defines hePointer as a smart pointer to a tHalfEdge object.
     */
    typedef tSmartPtr<dcel::tHalfEdge>   hePointer;


    /**
     * Defines bzPointer as a smart pointer to a tBezier object.
     */
    typedef tSmartPtr<tBezier>           bzPointer;


    /**
     * Method for creating an instance of this object.
     *
     * @param v: a pointer to a vertex.
     *
     */
    static Pointer NEW(const vtPointer& v)
    {
      tMyVTAttrib *dumbPtr = new tMyVTAttrib;
      Pointer smartPtr(dumbPtr);

      smartPtr->set_owner(v);

      smartPtr->compute_degree();

      smartPtr->set_normal(t3DVector(0,0,0));

      smartPtr->set_patch(bzPointer(0));

      return smartPtr;
    }


    /**
     * Destructor.
     */
    virtual ~tMyVTAttrib() { m_pat = bzPointer(0); }


    /**
     * Returns the degree  of the origin vertex of  the half-edge that
     * owns this attribute.
     *
     * @return The degree  of the origin vertex of  the half-edge that
     * owns this attribute.
     *
     */
    unsigned get_degree() const { return m_d; }


    /**
     * Returns the  unit normal associated  with the vertex  that owns
     * this attribute.
     *
     * @return The  unit normal associated  with the vertex  that owns
     * this attribute.
     *
     */
    t3DVector get_normal() const { return m_n; }


    /**
     * Return  a pointer  to the  rectangular Bezier  patch associated
     * with the half-edge that owns this attribute.
     *
     * @return a  pointer to  the rectangular Bezier  patch associated
     * with the half-edge that owns this attribute.
     *
     */
    bzPointer get_patch() const { return m_pat; }


    /**
     * Assigns a vector to the  unit normal associated with the vertex
     * that owns this attribute.
     *
     * @param A  vector to be  assigned to the unit  normal associated
     * with the vertex that owns this attribute.
     *
     */
    void set_normal(const t3DVector& n) { m_n = n; }


    /**
     * Assign a value  to the pointer to the  rectangular Bezier patch
     * associated with the half-edge that owns this attribute.
     *
     * @param b: a pointer to a rectangular Bezier patch.
     *
     */
    void set_patch(const bzPointer& b) { m_pat = b; }


  private:
    /**
     * Default constructor.
     */
    tMyVTAttrib() {}


    /**
     * Assigns  a value  to the  degree of  the origin  vertex  of the
     * half-edge that owns this attribute.
     *
     * @param An unsigned integer to  be assigned to the degree of the
     * origin vertex of the half-edge that owns this attribute.
     *
     */
    void set_degree(unsigned d) { m_d = d; }


    /**
     * Computes the degree of the  origin vertex of the half-edge that
     * owns this attribute.
     *
     */
    void compute_degree();


    /**
     * Degree of  the origin  vertex of the  half-edge that  owns this
     * attribute.
     */
    unsigned m_d;    


    /**
     * Unit normal associated with the vertex that owns this attribute.
     */
    t3DVector m_n;


    /**
     * A pointer  to the rectangular Bezier patch  associated with the
     * vertex that owns this attribute.
     */
    bzPointer m_pat;

  };


}


#endif   // MYVTATTRIB_H_
