/** 
 * @class pps::tPNTriangle
 *  
 * @brief A class that implements PN triangles.
 *
 * @author
 * Marcelo Ferreira Siqueira @n
 * Universidade Federal de Mato Grosso do Sul, @n
 * Departamento de Computacao e Estatistica, @n
 * marcelo at dct (dot) ufms (dot) br
 *
 * @version 1.0
 * @date Abril 2008
 *
 * @attention This program is distributed  WITHOUT ANY WARRANTY,  and
 *            it may be freely redistributed under the condition that
 *            the copyright notices are  not  removed,  and  no
 *            compensation  is  received. Private, research, and
 *            institutional use  is free. Distribution of this code as
 *            part of a commercial system is permissible ONLY BY DIRECT
 *            ARRANGEMENT WITH THE AUTHOR.
 */

#include "pntriangle.h"

#include <sstream>

#include "mathmacros.h"


namespace pps {


  /**
   * Computes the PN triangle coefficients
   *
   * @param p1: first vertex of the triangle defining the PN triangle.
   * @param p2: second vertex of the triangle defining the PN triangle.
   * @param p3: third vertex of the triangle defining the PN triangle.
   * @param n1: unit normal at the first vertex.
   * @param n2: unit normal at the second vertex.
   * @param n3: unit normal at the third vertex.
   * 
   */
  void
  tPNTriangle::build(
                     const t3DVector& p1,
                     const t3DVector& p2,
                     const t3DVector& p3,
                     const t3DVector& n1,
                     const t3DVector& n2,
                     const t3DVector& n3
		    )
    throw (tExceptionObject)
  {
    if ((ABS(1-n1.length()) >= 1e-12) || 
        (ABS(1-n2.length()) >= 1e-12) ||
        (ABS(1-n3.length()) >= 1e-12)) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "Face normal does not have unit length";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    m_b300 = p1;
    m_b030 = p2;
    m_b003 = p3;

    double w;

    w = (p2 - p1).inner(n1);
    m_b210 = (1 / 3.) * ( (2 * p1) + p2 - (w * n1) );

    w = (p1 - p2).inner(n2);
    m_b120 = (1 / 3.) * ( (2 * p2) + p1 - (w * n2) );

    w = (p3 - p2).inner(n2);
    m_b021 = (1 / 3.) * ( (2 * p2) + p3 - (w * n2) );

    w = (p2 - p3).inner(n3);
    m_b012 = (1 / 3.) * ( (2 * p3) + p2 - (w * n3) );

    w = (p1 - p3).inner(n3);
    m_b102 = (1 / 3.) * ( (2 * p3) + p1 - (w * n3) );

    w = (p3 - p1).inner(n1);
    m_b201 = (1 / 3.) * ( (2 * p1) + p3 - (w * n1) );

    t3DVector e((1/6.)*(m_b210+m_b120+m_b021+m_b012+m_b102+m_b201));
    t3DVector v((1/3.)*(p1 + p2 + p3));

    m_b111 = e + (0.5 * (e - v)); 


    m_n200 = n1;
    m_n020 = n2;
    m_n002 = n3;


    double d;

    t3DVector mv1(p2 - p1);
    t3DVector mv2(n1 + n2);

    w = mv1.inner(mv2);
    d = mv1.inner(mv1);
    m_n110 = mv2 - ((2 * w) / d) * (mv1);

    mv1 = p3 - p1;
    mv2 = n1 + n3;
    w = mv1.inner(mv2);
    d = mv1.inner(mv1);
    m_n101 = mv2 - ((2 * w) / d) * (mv1);

    mv1 = p3 - p2;
    mv2 = n2 + n3;
    w = mv1.inner(mv2);
    d = mv1.inner(mv1);
    m_n011 = mv2 - ((2 * w) / d) * (mv1);

    m_n110.normalize();
    m_n101.normalize();
    m_n011.normalize();
  }


  /**
   * Computes a point on this PN triangle.
   *
   * @param u: first barycentric coordinate of the given point.
   * @param v: second barycentric coordinate of the given point.
   * @param w: third barycentric coordinate of the given point.
   *
   */
  t3DVector
  tPNTriangle::point(
                     double u,
                     double v,
                     double w
		    )
    const throw (tExceptionObject)
  {

    if (ABS(u) <= 1e-15) {
      u = 0;
    }
    else if (ABS(1-u) <= 1e-15) {
      u = 1;
    }

    if (ABS(v) <= 1e-15) {
      v = 0;
    }
    else if (ABS(1-v) <= 1e-15) {
      v = 1;
    }

    if (ABS(w) <= 1e-15) {
      w = 0;
    }
    else if (ABS(1-w) <= 1e-15) {
      w = 1;
    }

    if ((u < 0) || (u > 1) || 
        (v < 0) || (v > 1) ||
        (w < 0) || (w > 1)) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "Barycentric coordinates are out of range";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    double u2 = u * u;
    double v2 = v * v;
    double w2 = w * w;

    double u3 = u * u2;
    double v3 = v * v2;
    double w3 = w * w2;

    t3DVector pt(u3 * m_b300);

    pt += (v3              * m_b030);
    pt += (w3              * m_b003);
    pt += ((3 * u2 * v)    * m_b210);
    pt += ((3 * u * v2)    * m_b120);
    pt += ((3 * u2 * w)    * m_b201);
    pt += ((3 * v2 * w)    * m_b021);
    pt += ((3 * u * w2)    * m_b102);
    pt += ((3 * v * w2)    * m_b012);
    pt += ((6 * u * v * w) * m_b111);

    return pt;
  }


  /**
   * Computes the unit normal of this PN triangle at a given point.
   *
   * @param u: first barycentric coordinate of the given point.
   * @param v: second barycentric coordinate of the given point.
   * @param w: third barycentric coordinate of the given point.
   *
   */
  t3DVector
  tPNTriangle::normal(
                      double u,
                      double v,
                      double w
		     )
    const throw (tExceptionObject)
  {

    if (ABS(u) <= 1e-15) {
      u = 0;
    }
    else if (ABS(1-u) <= 1e-15) {
      u = 1;
    }

    if (ABS(v) <= 1e-15) {
      v = 0;
    }
    else if (ABS(1-v) <= 1e-15) {
      v = 1;
    }

    if (ABS(w) <= 1e-15) {
      w = 0;
    }
    else if (ABS(1-w) <= 1e-15) {
      w = 1;
    }

    if ((u < 0) || (u > 1) || 
        (v < 0) || (v > 1) ||
        (w < 0) || (w > 1)) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "Barycentric coordinates are out of range";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    double u2 = u * u;
    double v2 = v * v;
    double w2 = w * w;

    t3DVector pt(
      ((u2 * m_n200) + (v2 * m_n020) + (w2 * m_n002) +
       ((u * v) * m_n110) + ((u * w) * m_n101) + 
       ((v * w) * m_n011)));

    return pt.normalize();
  }


}
